indexing
	description: "Program points: points within Eiffel classes, simple (correspond to single position in source code) or %
		%aggregative (correspong to set of positions), characterized by kinds of variables in scope %
		%and kind of assertions that are checked at it"
	author: ""
	date: "$Date$"
	revision: "$Revision$"

deferred class
	CI_PROGRAM_POINT

inherit
	CI_CODE_GENERATION
		export {NONE}
			all
		end

	CI_OPERATOR_IDENTIFIERS
		export {NONE}
			all
		end

	CI_SETTINGS_CLIENT

feature -- Initialization
	make (a_class: like surrounding_class; a_universe: like universe) is
			-- Create a program point within `a_class' in `a_universe'
		require
			a_class_exists: a_class /= Void
			a_universe_exists: a_universe /= Void
		do
			surrounding_class := a_class
			universe := a_universe
			create {TWO_WAY_SORTED_SET [CI_VARIABLE]} variables.make
			variables.compare_objects
			create {LINKED_LIST [CI_INFERRED_ASSERTION]} inferred_assertions.make
			inferred_assertions.compare_objects
		ensure
			surrounding_class_set: surrounding_class = a_class
			universe_set: universe = a_universe
		end

feature -- Access
	surrounding_class: ET_CLASS
			-- Class to which program point belongs

	universe: ET_UNIVERSE
			-- Universe to which program point belongs			

	suppressor: CI_PROGRAM_POINT is
			-- Program point, who's assertions suppress assertions at current program point (Void if no suppressor)
		do
			-- Default value is Void
		end

feature -- Status report
	ensures_invariant: BOOLEAN is
			-- Should invariant of instrumented class hold at program point?
		do
			-- By default answer in "No"
		end

feature {CI_PROGRAM_POINT_PRINTER} -- Printing
	kind (a_visitor: CI_PROGRAM_POINT_PRINTER): STRING is
			-- Program point kind name (for visiting with printer)
		require
			a_visitor_exists: a_visitor /= Void
		deferred
		ensure
			result_exists: Result /= Void
		end

feature -- Variables
	variables: LINEAR_SUBSET [CI_VARIABLE]
			-- Variables in scope

	variables_added: BOOLEAN
			-- Are all variables in scope stored in `variables'?

	add_variables is
			-- Find variables in scope and store them into `variables'
		require
			variables_exists: variables /= Void
		deferred
		ensure
			variables_added: variables_added
		end

	named_variable (a_name: STRING): CI_VARIABLE is
			-- Variable with name `a_name'
		require
			a_name_exists: a_name /= Void
		do
			from
				variables.start
			until
				Result /= Void or variables.after
			loop
				if variables.item.name.is_equal (a_name) then
					Result := variables.item
				end
				variables.forth
			end
		end

feature -- Assertions
	checked_assertions: ET_ASSERTIONS is
			-- Eiffel assertions, that are checked at this program point
		deferred
		end

	state_printing_function (printing_instructions: ET_COMPOUND; locals: ET_LOCAL_VARIABLE_LIST): CI_STATE_PRINTING_FUNCTION is
			-- Function to be called from `checked_assertions' in instrumented system, that prints visible state at this program point
			-- using `printing_instructions', needs `locals' as locals and returns True
		require
			printing_instructions_exists: printing_instructions /= Void
		local
			compound: ET_COMPOUND
		do
			create Result.make (state_printing_function_name, state_printing_function_formals, type_from_nongeneric_class (universe.boolean_class),
				surrounding_class)
			if locals /= Void and then not locals.is_empty then
				Result.set_locals (locals)
			end
			create compound.make_with_capacity (printing_instructions.count + 1)
			compound.put_first (create {ET_ASSIGNMENT}.make (tokens.result_keyword, tokens.true_keyword))
			insert_compound_to_front (compound, printing_instructions)
			Result.set_compound (compound)
		ensure
			result_exists: Result /= Void
		end

	instrument is
			-- Instrument program point by making `checked_assertions' call `state_printing_function'
		do
			create_assertions
			checked_assertions.wipe_out
			checked_assertions.force_first (create {ET_CALL_EXPRESSION}.make (Void, state_printing_function_name, state_printing_function_actuals))
		end

	inferred_assertions: LIST [CI_INFERRED_ASSERTION]
			-- Assertions that were inferred for this program point

	annotate is
			-- Annotate program point with `inferred_assertions'
		do
			create_assertions
			process_original_assertions
			from
				inferred_assertions.finish
			until
				inferred_assertions.before
			loop
				checked_assertions.force_first (inferred_assertions.item)
				inferred_assertions.back
			end
		end

feature {NONE} -- Assertions
	create_assertions is
			-- Create empty `checked_assertions', if they don't exist
		deferred
		ensure
			checked_assertions_exists: checked_assertions /= Void
		end

	process_original_assertions is
			-- Process original assertions according to settings while annotating this point
		require
			checked_assertions_exists: checked_assertions /= Void
		deferred
		end

	state_printing_function_name: ET_FEATURE_NAME is
			-- Name of `state_printing_function' for this program point
		deferred
		end

	state_printing_function_formals: ET_FORMAL_ARGUMENT_LIST is
			-- Formal arguments of `state_printing_function'
		deferred
		end

	state_printing_function_actuals: ET_ACTUAL_ARGUMENT_LIST is
			-- Actual arguments of `state_printing_function' call inside instrumented `checked_assertions'
		deferred
		end

invariant
	surrounding_class_exits: surrounding_class /= Void
	universe_exists: universe /= Void
	variables_exists: variables /= Void
	variables_compared_by_value: variables.object_comparison
	inferred_assertions_exists: inferred_assertions /= Void
	state_printing_function_name_exists: state_printing_function_name /= Void
end
