indexing
	description: "Daikon proxies (objects that communicate with Daikon Invariant Detector)"
	author: ""
	date: "$Date$"
	revision: "$Revision$"

class
	CI_DAIKON

inherit
	CI_SETTINGS_CLIENT

	CI_ERROR_HANDLER_CLIENT

	EXECUTION_ENVIRONMENT
		export {NONE}
			all
		end

	KL_SHARED_FILE_SYSTEM

	PROCESS_FACTORY
		export {NONE}
			all
		end

create
	make_null

feature -- Basic operations
	check_environment_variables is
			-- Check, whether needed environment variables are set and set additional environmant variables
		do
			check_daikon_dir
			if daikon_dir /= Void then
				check_java_variables
				if jdk_dir /= Void or jre_dir /= Void then
					if file_system.file_exists (java_executable) then
						set_class_path
						environment_variables_ok := True
					else
						error_handler.report_no_java (java_bin)
					end
				end
			end
		end

	execute is
			-- Execute invariant detector
		require
			environment_variables_ok: environment_variables_ok
		do
			daikon_process := process_launcher_with_command_line (daikon_command_line, Settings.daikon_files_path)
			daikon_process.redirect_output_to_agent (agent error_handler.report_info_message)
			daikon_process.redirect_error_to_same_as_output
			daikon_process.set_buffer_size (512)
			daikon_process.enable_launch_in_new_process_group
			daikon_process.set_on_fail_launch_handler (agent error_handler.report_launch_failed (java_executable))
			daikon_process.set_hidden (True)
			daikon_process.launch
			if daikon_process.launched then
				daikon_process.wait_for_exit
			end
			if daikon_process.has_exited and daikon_process.exit_code /= 0 then
				error_handler.report_daikon_execution_failed
			end
		end

	print_invariants is
			-- Execute Print Invariants tool
		require
			environment_variables_ok: environment_variables_ok
			daikon_output_exists: file_system.file_exists (Settings.daikon_output_name)
		local
			out_file: KI_TEXT_OUTPUT_FILE
		do
			out_file := file_system.new_output_file (Settings.invariants_name)
			out_file.open_write
			if out_file.is_open_write then
				print_invariants_process := process_launcher_with_command_line (print_invariants_command_line, Settings.daikon_files_path)
				print_invariants_process.redirect_output_to_agent (agent out_file.put_string)
				print_invariants_process.redirect_error_to_agent (agent error_handler.report_info_message)
				print_invariants_process.set_buffer_size (512)
				print_invariants_process.enable_launch_in_new_process_group
				print_invariants_process.set_on_fail_launch_handler (agent error_handler.report_launch_failed (java_executable))
				print_invariants_process.set_hidden (True)
				print_invariants_process.launch
				if print_invariants_process.launched then
					print_invariants_process.wait_for_exit
				end
				if print_invariants_process.has_exited and print_invariants_process.exit_code /= 0 then
					error_handler.report_print_invariants_execution_failed
				end
				out_file.close
			else
				error_handler.report_error_message ("Can't create file " + Settings.invariants_name)
			end
		end

feature -- Status report
	environment_variables_ok: BOOLEAN
			-- Are needed environment variables set?

feature -- Environment variables
	daikon_dir_name: STRING is "DAIKONDIR"
			-- Name of "DAIKONDIR" environment variable

	daikon_dir: DIRECTORY_NAME
			-- Value of "DAIKONDIR" environment variable (path to Daikon)

	jdk_dir_name: STRING is "JDKDIR"
			-- Name of "JDKDIR" environment variable

	jdk_dir: DIRECTORY_NAME
			-- Value of "JDKDIR" environment variable (path to Java Development Kit)

	jre_dir_name: STRING is "JREDIR"
			-- Name of "JREDIR" environment_variable

	jre_dir: DIRECTORY_NAME
			-- Value of "JREDIR" environment variable (path to Java Runtime Environment)	

	class_path_name: STRING is "CLASSPATH"
			-- Name of "CLASSPATH" environment_variable	

	path_separator: CHARACTER is ';'
			-- Path separator for environment variables containing multiple paths

feature {NONE} -- Implementation
	daikon_process: PROCESS
			-- Process, in which Daikon is lauched

	print_invariants_process: PROCESS
			-- process, in which Print Invariants is launched

	daikon_command_line: STRING is
			-- Command line to launch Daikon
		once
			Result := java_executable.string
			Result.append (" daikon.Daikon --no_text_output ")
			Result.append (Settings.daikon_runtime_input_name)
			Result.append_character (' ')
			Result.append (Settings.daikon_static_input_name)
			if Settings.daikon_arguments /= Void then
				Result.append_character (' ')
				Result.append (Settings.daikon_arguments)
			end
		end

	print_invariants_command_line: STRING is
			-- Command line to launch Print Invariants
		once
			Result := java_executable.string
			Result.append (" daikon.PrintInvariants --format Eiffel --config_option daikon.PrintInvariants.remove_post_vars=1 ")
			Result.append (Settings.daikon_output_name)
		end

	java_bin: DIRECTORY_NAME is
			-- Path to java binaries
		once
			create Result.make
			if jdk_dir /= Void then
				Result.set_directory (jdk_dir)
			else
				Result.set_directory (jre_dir)
			end
			Result.set_subdirectory ("bin")
		end

	java_executable: FILE_NAME is
			-- Name of JVM executable file
		once
			create Result.make
			Result.set_directory (java_bin)
			Result.set_file_name ("java")
			if not Settings.exe_extension.is_empty then
				Result.add_extension (Settings.exe_extension)
			end
		end

	check_daikon_dir is
			-- Check, whether "DAIKONDIR" is set properly
		local
			str: STRING
		do
			str := get (daikon_dir_name)
			if str = Void then
				error_handler.report_daikon_dir_not_set (daikon_dir_name)
			else
				create daikon_dir.make_from_string (str)
				if not file_system.directory_exists (daikon_dir) then
					error_handler.report_nonexistent_daikon_dir (daikon_dir_name, daikon_dir)
					daikon_dir := Void
				end
			end
		end

	check_java_variables is
			-- Check, whether "JDKDIR" or "JREDIR" is set properly
		local
			str: STRING
		do
			str := get (jdk_dir_name)
			if str = Void then
				str := get (jre_dir_name)
				if str = Void then
					error_handler.report_java_variables_not_set (jdk_dir_name, jre_dir_name)
				else
					create jre_dir.make_from_string (str)
					if not file_system.directory_exists (jre_dir) then
						error_handler.report_nonexistent_java_dir (jre_dir_name, jre_dir)
						jre_dir := Void
					end
				end
			else
				create jdk_dir.make_from_string (str)
				if not file_system.directory_exists (jdk_dir) then
					error_handler.report_nonexistent_java_dir (jdk_dir_name, jdk_dir)
					jdk_dir := Void
				end
			end
		end

	set_class_path is
			-- Set appropriate value for "CLASSPATH"
		require
			daikon_dir_exists: daikon_dir /= Void
			java_dir_exists: jdk_dir /= Void or jre_dir /= Void
		local
			class_path: STRING
			cpadd: FILE_NAME
		do
			create cpadd.make_from_string (daikon_dir)
			cpadd.set_file_name ("daikon.jar")
			class_path := get (class_path_name)
			if class_path = Void or else class_path.is_empty then
				class_path := cpadd.string
			else
				class_path.append_character (path_separator)
				class_path.append (cpadd)
			end

			if jdk_dir /= Void then
				create cpadd.make_from_string (jdk_dir)
				cpadd.set_subdirectory ("jre")
				cpadd.set_subdirectory ("lib")
				cpadd.set_file_name ("rt.jar")
				class_path.append_character (path_separator)
				class_path.append (cpadd)
			else
				check jre_dir /= Void end
				create cpadd.make_from_string (jre_dir)
				cpadd.set_subdirectory ("lib")
				cpadd.set_file_name ("rt.jar")
				class_path.append_character (path_separator)
				class_path.append (cpadd)
			end

			put (class_path, class_path_name)
		end

invariant
	daikon_command_line_not_empty: not daikon_command_line.is_empty
end
