indexing
	description: "Trackable variables (expressions that describe visible program state at some program point)"
	author: ""
	date: "$Date$"
	revision: "$Revision$"

deferred class
	CI_VARIABLE

inherit
	COMPARABLE

	CI_SETTINGS_CLIENT
		undefine
			is_equal	-- Variables are compared by `name'
		end

	CI_IDENTITIES_NAMES_CLIENT
		undefine
			is_equal	-- Variables are compared by `name'
		end

	CI_RUNTIME_OUTPUT_NAMES_CLIENT
		undefine
			is_equal	-- Variables are compared by `name'
		end

	CI_CODE_GENERATION
		export {NONE}
			all
		undefine
			is_equal	-- Variables are compared by `name'
		end

feature -- Access
	name: STRING is
			-- Variable name in Daikon format (unique within program point)
		deferred
		end

	eiffel_name: STRING is
			-- Variable name, as it will appear in annotated class
		do
			Result := name
		end

	type: CI_VARIABLE_TYPE
			-- Declared type

	surrounding_routine: ET_ROUTINE
			-- Routine, to which variable's program point belongs
			-- Void if variable is from class-wide program point like stable time

	clients: ET_CLASS_NAME_LIST
			-- Classes that can observe variable depending on program point, where it appears
			-- (empty list is equivalent to {NONE})

	constant: ET_CONSTANT is
			-- Compile-time constant value of variable (Void if not a constant)
		do
			-- By default value is Void	
		end

feature -- Evaluation
	as_expression: ET_EXPRESSION is
			-- Variable viewed as Eiffel expression; to be used as target for queries
		deferred
		end

	identity: ET_EXPRESSION is
			-- Expression with printable result that identifies variable value
			-- For printable variables `identity' = `as_expression'
			-- For reference variables `identity' = `($as_expression).hash_code'
			-- For expanded variables `identity' is a function of its attributes' identities
		do
			Result := type.kind.identity (Current)
		end

	synonim: ET_IDENTIFIER
			-- Identifier of local variable that has same value as `as_expression'

	evaluation_preconditions: ET_PRECONDITIONS
			-- List of conditions, which define when expression `as_expression' can be evaluated
			-- If list is empty, expression has trivial precondition `True'

	absolute_preconditions: ET_PRECONDITIONS is
			-- Absolute list of preconditions, which includes also `evaluation_preconditions' of all parent variables
		do
			Result := evaluation_preconditions
		end

feature {CI_VARIABLE_TYPE_KIND} -- Evaluation
	address_expression: ET_ADDRESS_EXPRESSION is
			-- Variable address; may be used in identity expression
		deferred
		end

feature -- Status report
	needs_synonim: BOOLEAN is
			-- Does variable need a synonim or `as_expression' can be used everywhere as is?
		once
			-- By default answer is "No"			
		end

	is_current: BOOLEAN is
			-- Is variable a "Current" entity
		once
			-- By default answer is "No"
		end

	is_query: BOOLEAN is
			-- Is variable a query?
		once
			-- By default answer is "No"
		end

	is_param: BOOLEAN is
			-- Daikon property: variable value can't be changed by routine
		once
			-- By default answer is "No"
		end

	was_nonsensical: BOOLEAN
			-- Was the variable nonsensical at least once during system execution?

	instrumentation_mode_disabled: BOOLEAN
			-- Is variable not a part of instrumented system?

	has_watched_once: BOOLEAN is
			-- Does variable contain a watched once function?
		once
			-- By default answer is "No"
		end

feature -- Status setting
	set_was_nonsensical (b: BOOLEAN) is
			-- Set `was_nonsensical' to `b'
		do
			was_nonsensical := b
		ensure
			was_nonsensical_set: was_nonsensical = b
		end

	set_instrumentation_mode (b: BOOLEAN) is
			-- Set `instrumentation_mode_disabled' to `not b'
		do
			instrumentation_mode_disabled := not b
		ensure
			instrumentation_mode_set: instrumentation_mode_disabled = not b
		end

feature -- Unfolding
	unfolding_depth: INTEGER
			-- Depth to which unfolding is done	(should be initialized by descendants)	

	set_unfolding_depth (a_depth: INTEGER) is
			-- Set `unfolding_depth' to `a_depth'
		require
			a_depth_not_negative: a_depth >= 0
		do
			unfolding_depth := a_depth
		ensure
			unfolding_depth_set: unfolding_depth = a_depth
		end

	default_unfolding_depth: INTEGER is 1
			-- Default value of `unfolding_depth'

	children: LIST [CI_VARIABLE] is
			-- Child variables of this variables
		do
			Result := type.kind.children (Current)
		end

	unfolded: TWO_WAY_TREE [CI_VARIABLE] is
			-- An unfolded representation of variable
			-- Child nodes are all zero-argument queries exported to `clients'
			-- Special handling needed for containers (not implemented yet)
		do
			create Result.make (Current)
			unfold (Result, unfolding_depth)
		end

feature -- Comparison
	infix "<" (other: like Current): BOOLEAN is
			-- Is current object less than `other'?
		do
			Result := (name < other.name)
		end

feature {NONE} -- Implementation
	unfold (a_node: TWO_WAY_TREE [CI_VARIABLE]; i: INTEGER) is
			-- If depth is not reached (i > 0), find children of `a_node'
			-- and call `unfold' for each of them
		require
			a_node_exists: a_node /= Void
			a_node_has_item: a_node.item /= Void
			i_not_negative: i >= 0
			i_not_too_large: i <= unfolding_depth
		local
			var: CI_VARIABLE
			c: LIST [CI_VARIABLE]
			child_node: TWO_WAY_TREE [CI_VARIABLE]
		do
			if i > 0 then
				-- Unfolding depth is not reached
				var := a_node.item

				-- Add children of `var' to `a_node' children
				c := var.children
				from
					c.start
				until
					c.after
				loop
					create child_node.make (c.item)
					a_node.put_child (child_node)
					c.forth
				end

				-- Call `unfold' for children of `a_node'
				from
					a_node.child_start
				until
					a_node.child_after
				loop
					unfold (a_node.child, i - 1)
					a_node.child_forth
				end
			end
		end

	is_sufficiently_exported (a_feature: ET_FEATURE): BOOLEAN is
			-- Is `a_feature' exported to all `clients'?
		require
			a_feature_exists: a_feature /= Void
		local
			i: INTEGER
		do
			from
				i := 1
				Result := True
			until
				i > clients.count
			loop
				if not a_feature.is_exported_to (type.universe.eiffel_class (clients.class_name (i)), type.universe) then
					Result := False
					i := clients.count	-- Jump out of the loop
				end
				i := i + 1
			end
		end

feature {CI_VARIABLE_TYPE_KIND} -- Implementation
	use_query (a_query: ET_QUERY): BOOLEAN is
			-- Use `a_query' in unfolding `Current'?
		local
			function: ET_FUNCTION
			state_printing_function: CI_STATE_PRINTING_FUNCTION
		do
			function ?= a_query
			state_printing_function ?= a_query
			Result := Settings.is_valid_query_variable (a_query) and			-- satisfies current instrumentation settings
				state_printing_function = Void and 								-- not a state printing function
				(function /= Void implies function /= surrounding_routine) and  -- not currently instrumented function
				not Identities_names.is_identity_target (a_query.name,
					type.base_class.name) and									-- not identity target		
				not RO_names.is_trace_feature_name (a_query.name) and			-- not trace feature
				is_sufficiently_exported (a_query)								-- is exported to all `clients'
		end

feature {CI_VARIABLE} -- Implementation
	synonim_name: STRING is
			-- Name of `synonim'
		do
			create Result.make (name.count + 1)
			Result.append (name.as_lower)
			Result.append_character ('_')
		end

invariant
	name_exists: name /= Void
	name_not_empty: not name.is_empty
	eiffel_name_exists: eiffel_name /= Void
	eiffel_name_not_empty: not eiffel_name.is_empty
	type_exist: type /= Void
	clients_exist: clients /= Void
	unfolded_exist: unfolded /= Void
	unfolding_depth_not_negative: unfolding_depth >= 0
	as_expression_exists: as_expression /= Void
	address_expression_exists_when_needed: type.kind.uses_address implies address_expression /= Void
	identity_exists: identity /= Void
	evaluation_preconditions_exists: evaluation_preconditions /= Void
	absolute_preconditions_exists: absolute_preconditions /= Void
	synonim_name_exists: synonim_name /= Void
--	synonim_name_is_valid_identifier	
	synonim_exists_when_needed: needs_synonim implies synonim /= Void
	synonim_has_right_name: synonim /= Void implies synonim.name = synonim_name
end
